import time
import json

import logging

logger = logging.getLogger(__name__)

class AuthToken():
    def __init__(
            self, 
            id_token: str, 
            access_token: str, 
            expires_in: int, 
            token_type: str, 
            refresh_token: str, 
            scope: str,
            expiration_time=None):
        
        logger.info(f"Create AuthToken.")

        self.id_token = id_token
        self.access_token = access_token
        self.expires_in = expires_in
        self.token_type = token_type
        self.refresh_token = refresh_token
        self.scope = scope

        self.expiration_time = expiration_time
        if not expiration_time:
            self.expiration_time = time.time() + expires_in
        logger.info(f"AuthToken expires at {self.expiration_time} seconds.")
    
    def is_expired(self):
        expired = time.time() > self.expiration_time
        logger.info(f"Check if auth token is expired. Expired status is {expired}.")
        return expired

    def save_json(self, token_file_path):
        logger.info(f"Save auth token to file {token_file_path}.")
        json.dump(self.__dict__, open(token_file_path, "w"), indent=4)
        



import datetime
import pytz
from timezonefinder import TimezoneFinder
from typing import List
import logging

from commutauto.exceptions import ExpiredSearchException

logger = logging.getLogger(__name__)

def check_datetime_utc(date: datetime.datetime):
    if date.tzinfo != pytz.utc:
        raise ValueError(f"Date '{date}' must be expressed in UTC time.")


def infer_local_timezone(latitude: float, longitude: float) -> datetime.tzinfo:
    tf = TimezoneFinder()
    timezone_str = tf.timezone_at(lng=longitude, lat=latitude)
    tz = pytz.timezone(timezone_str)
    logger.info(f"Inferred timezone: {tz}.")
    return tz

def convert_datetime_to_utc(date: datetime.datetime, latitude: float, longitude: float) -> datetime.datetime:
    tz = infer_local_timezone(latitude, longitude)
    utc_date = tz.localize(date).astimezone(pytz.utc)
    logger.info(f"Date {date} converted to UTC {utc_date}.")
    return utc_date

def round_to_quarter_hour(date):
    """Round datetime to nearest quarter-hour"""
    logger.info(f"Rounding date {date} to closest quarter hour.")

    quarter_ratio = date.minute / 15 - date.minute // 15
    round_up = int(quarter_ratio > 0.5)
    quarter_hour = ((date.minute // 15) + round_up) * 15
    if quarter_hour > 45:
        date += datetime.timedelta(hours=1)
    date = date.replace(minute=quarter_hour % 60)
    logger.info(f"Rounded date: {date}.")
    date = date.replace(second=0, microsecond=0)
    return date

def round_booking_time_window(start_date: datetime.datetime, end_date: datetime.datetime):
    start_date = round_to_quarter_hour(start_date)
    end_date = round_to_quarter_hour(end_date)
    return start_date, end_date

def validate_booking_request_time_window(start_date: datetime.datetime, end_date: datetime.datetime):
    if end_date <= start_date:
        raise ValueError(f"End date '{end_date}' must be later than start date '{start_date}'.")
    
    if end_date < start_date+ datetime.timedelta(minutes=30):
        raise ValueError(f"Booking duration must be at least 30 minutes. Got {end_date-start_date}")

def check_start_date_greater_than_quarter_hour_from_now(start_date):
    if datetime.datetime.now(pytz.utc) + datetime.timedelta(minutes=15) > start_date:
        raise ExpiredSearchException(f"Booking search start date '{start_date}' is in less than 15 minutes from now.")

def check_start_date_sooner_than_one_month(start_date):
    cur_date = datetime.datetime.now(pytz.utc)
    cur_date = cur_date.replace(hour = 0, minute = 0, second =0, microsecond=0)

    # round hour to 0am then add 31 days
    if cur_date + datetime.timedelta(days=32) < start_date:
        raise ExpiredSearchException(f"Booking search start date '{start_date}' is in more than 31 days from now.")
    
def check_no_booking_intersection_with_search_window(start_date: datetime.datetime, end_date: datetime.datetime, reservations: List) -> List:
    overlapping_reservations = []
    for res in reservations:
        interval1 = [start_date, end_date]
        interval2 = [res.StartDateUTC, res.EndDateUTC]

        if interval2[0] < interval1[0]:
            temp = interval2
            interval2 = interval1
            interval1 = temp

        if interval2[0] < interval1[1]:
            overlapping_reservations.append(res)
    return overlapping_reservations


from abc import ABC, abstractmethod
import datetime
import logging

from commutauto.constants import AspScript, LSIRoute
from commutauto.booking_service.date_manager import check_datetime_utc, infer_local_timezone

logger = logging.getLogger(__name__)

class Payload(ABC):
    @abstractmethod
    def to_dict(self) -> dict:
        raise NotImplementedError
    
    def is_url_encoded(self) -> bool:
        raise NotImplementedError

class GetCarDisponibilityPayload(Payload):
    def __init__(
            self, 
            StartDateUTC: datetime.datetime, # UTC timezone
            EndDateUTC: datetime.datetime, 
            Latitude: float, 
            Longitude: float, 
            CityID: int,
            CurrentLanguageID: int = 2, 
            Accessories: str = 0, 
            FeeType: str = None):
        
        check_datetime_utc(StartDateUTC)
        check_datetime_utc(EndDateUTC)

        self._StartDateUTC = StartDateUTC
        self._EndDateUTC = EndDateUTC
        self.Latitude = Latitude
        self.Longitude = Longitude
        self.CurrentLanguageID = CurrentLanguageID
        self.CityID = CityID
        self.Accessories = Accessories
        self.FeeType = FeeType

        self._LocalTimeZone = infer_local_timezone(self.Latitude, self.Longitude)

    def to_dict(self):
        d = {key: value for key, value in self.__dict__.items() if key[0] != "_"} # TODO: infer out timezone from coordinates
        d["StartDate"] = self._StartDateUTC.astimezone(self._LocalTimeZone).strftime('%d/%m/%Y %H:%M')
        d["EndDate"] = self._EndDateUTC.astimezone(self._LocalTimeZone).strftime('%d/%m/%Y %H:%M')
        d["StartDateISO"] = self._StartDateUTC.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        d["EndDateISO"] = self._EndDateUTC.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return d
    
    def is_url_encoded(self) -> bool:
        return True

class BookCarPayload(Payload):
    def __init__(
        self, 
        StartDateUTC: datetime.datetime, # UTC timezone
        EndDateUTC: datetime.datetime,
        CarId: int, 
        Latitude: float, 
        Longitude: float,
        CityID: int,
        CurrentLanguageID: int = 2, 
        ReservationCityID: str = "undefined",
        Accessories: int = 0,
        FeeType: str = None,
        DestinationId: int = None,
        Note: str = "",
        ExpectedNbrKm: str = None,
        AcceptAdditionalCharges: str = "undefined"
    ):
        
        check_datetime_utc(StartDateUTC)
        check_datetime_utc(EndDateUTC)

        self._StartDateUTC = StartDateUTC
        self._EndDateUTC = EndDateUTC
        self.CarId = CarId
        self.CurrentLanguageID = CurrentLanguageID
        self.CityID = CityID
        self.ReservationCityID = ReservationCityID
        self.Accessories = Accessories
        self.FeeType = FeeType
        self.DestinationId = DestinationId
        self.Note = Note
        self.ExpectedNbrKm = ExpectedNbrKm
        self.AcceptAdditionalCharges = AcceptAdditionalCharges

        self._LocalTimeZone = infer_local_timezone(Latitude, Longitude)


    def to_dict(self):
        d = {key: value for key, value in self.__dict__.items() if key[0] != "_"}
        d["StartDate"] = self._StartDateUTC.astimezone(self._LocalTimeZone).strftime('%d/%m/%Y')
        d["EndDate"] = self._EndDateUTC.astimezone(self._LocalTimeZone).strftime('%d/%m/%Y')
        d["StartTime"] = self._StartDateUTC.astimezone(self._LocalTimeZone).strftime('%H:%M')
        d["EndTime"] = self._EndDateUTC.astimezone(self._LocalTimeZone).strftime('%H:%M')
        return d
    
    def is_url_encoded(self) -> bool:
        return True

class GetBookingPayload(Payload):
    def __init__(
        self,
        ReservationID: int,
        CurrentLanguageID: int = 2, 
    ):
        self.ReservationID = ReservationID
        self.CurrentLanguageID = CurrentLanguageID

    def to_dict(self):
        return self.__dict__
    
    def is_url_encoded(self) -> bool:
        return True

class CancelBookingPayload(Payload):
    def __init__(self,
                    StartDateUTC: datetime.datetime,
                    EndDateUTC: datetime.datetime,
                    ReservationID: int,
                    NewFeeType: int,
                    ExpectedNbrKm: int, 
                    Latitude: float, 
                    Longitude: float,
                    CurrentLanguageID: int = 2,
                    Action: str = "Supprimer",
                    NewDestinationID: str = "",
                    NewNote: str = ""):
        
        check_datetime_utc(StartDateUTC)
        check_datetime_utc(EndDateUTC)
        
        self._StartDateUTC = StartDateUTC
        self._EndDateUTC = EndDateUTC
        self.ReservationID = ReservationID
        self.NewFeeType = NewFeeType
        self.ExpectedNbrKm = ExpectedNbrKm
        self.NewDestinationID = NewDestinationID
        self.NewNote = NewNote
        self.Action = Action
        self.CurrentLanguageID = CurrentLanguageID

        self._LocalTimeZone = infer_local_timezone(Latitude, Longitude)


    def to_dict(self):
        d = {key: value for key, value in self.__dict__.items() if key[0] != "_"}
        d["StartDate"] = self._StartDateUTC.astimezone(self._LocalTimeZone).strftime('%d/%m/%Y')
        d["EndDate"] = self._EndDateUTC.astimezone(self._LocalTimeZone).strftime('%d/%m/%Y')
        d["StartTime"] = self._StartDateUTC.astimezone(self._LocalTimeZone).strftime('%H:%M')
        d["EndTime"] = self._EndDateUTC.astimezone(self._LocalTimeZone).strftime('%H:%M')
        return d
    
    def is_url_encoded(self) -> bool:
        return True

class GetBookingListPayload(Payload):
    def __init__(
        self,
        CurrentLanguageID: int = 2,
        Type: int = 1
    ):
        self.Type = Type
        self.CurrentLanguageID = CurrentLanguageID

    def to_dict(self):
        return self.__dict__
    
    def is_url_encoded(self) -> bool:
        return True
    
class GetFlexDisponibilityUrlParameter(Payload):
    def __init__(
            self, 
            BranchID: int,
            CityID: int,
            LanguageID: int = 2
        ):
        self.BranchID = BranchID
        self.CityID = CityID
        self.LanguageID = LanguageID
    def to_dict(self):
        return self.__dict__
    
    def is_url_encoded(self) -> bool:
        return True
    
class BookFlexPayload(Payload):
    def __init__(
            self,
            CarVIN: str,
            BranchID: int
        ):
        self.BranchID = BranchID
        self.CarVIN = CarVIN
    def to_dict(self):
        return self.__dict__
    
    def is_url_encoded(self) -> bool:
        return False
    
class CancelFlexBookingPayload(Payload):
    def __init__(
            self, 
            BranchID: int
        ):
        self.BranchID = BranchID
    def to_dict(self):
        return self.__dict__
    
    def is_url_encoded(self) -> bool:
        return False

class GetCurrentFlex(Payload):
    def __init__(
            self, 
            BranchID: int
        ):
        self.BranchID = BranchID
    def to_dict(self):
        return self.__dict__
    
    def is_url_encoded(self) -> bool:
        return False

PAYLOAD_FACTORY = {
    AspScript.GET_CAR_DISPONIBILITY: GetCarDisponibilityPayload,
    AspScript.BOOK_CAR: BookCarPayload,
    AspScript.GET_BOOKING: GetBookingPayload,
    AspScript.CANCEL_BOOKING: CancelBookingPayload,
    AspScript.GET_BOOKING_LIST: GetBookingListPayload,
    LSIRoute.GET_FLEX_DISPONIBILITY: GetFlexDisponibilityUrlParameter,
    LSIRoute.BOOK_FLEX: BookFlexPayload,
    LSIRoute.CANCEL_BOOKING_FLEX: CancelFlexBookingPayload,
    LSIRoute.GET_CURRENT_FLEX: GetCurrentFlex
}

def get_payload(script_name: str, payload_args: dict) -> Payload:
    return PAYLOAD_FACTORY[script_name](**payload_args)


from dataclasses import dataclass
from typing import List
import json
import logging

logger = logging.getLogger(__name__)

@dataclass
class Flex:
    CarId: int
    CarVin: str
    CarPlate: str
    CarModel: str
    CarNo: int
    Latitude: float
    Longitude: float
    EnergyLevel: int
    LastUseDate: str
    LastUse: int
    isPromo: bool
    BoardComputerType: int
    BookingStatus: int
    CarBrand: str
    CarColor: str
    CarSeatNb: int
    CarAccessories: List[int]
    IsElectric: bool
    VehiclePromotions: int
    CityID: int
    Distance: int = -1

    def __eq__(self, other):
        return self.Distance == other.Distance
        
    def __lt__(self, other):
        return self.Distance < other.Distance
        
    def __gt__(self, other):
        return self.Distance > other.Distance

    def __str__(self):
        return json.dumps(self.__dict__, indent=4)



from enum import Enum
from typing import Union

MINIMUM_SEARCH_RADIUS_KM = 0.3
# NOTE: shorter wait time for flex than for stations
# as the search is more time critical
FLEX_BASE_WAIT_TIME_SECOND=45 # 45 seconds
FLEX_WAIT_TIME_JITTER_SECOND=30 # 0 seconds
FLEX_EXPIRATION_TIME_SECOND=1*60*60 # 1 hour
MAX_FLEX_EXPIRATION_TIME_SECOND=3*60*60 # No more than 3 hours

STATION_BASE_WAIT_TIME_SECOND=2*60 # 2 minutes
STATION_WAIT_TIME_JITTER_SECOND=3*60 # 3 minutes
STATION_EXPIRATION_TIME_SECOND=1*24*60*60 # 1 day
MAX_STATION_EXPIRATION_TIME_SECOND=7*24*60*60 # No more than a week

class CommutautoEndpoint(Enum):
    ASP = "https://www.reservauto.net/Scripts/Client/Ajax/MobileApplication/Communauto/"
    LSI = "https://www.reservauto.net/WCF/LSI/LSIBookingServiceV3.svc/"

class AspScript(Enum):
    GET_CAR_DISPONIBILITY= "Get_Car_DisponibilityJSON.asp"
    BOOK_CAR = "ReservationAdd.asp"
    GET_BOOKING = "Get_Reservation.asp"
    GET_BOOKING_LIST = "ReservationList.asp"
    CANCEL_BOOKING= "ReservationModify.asp"

class LSIRoute(Enum):
    GET_FLEX_DISPONIBILITY= "GetAvailableVehicles"
    BOOK_FLEX = "CreateBookingPost"
    CANCEL_BOOKING_FLEX= "CancelBookingPost"
    GET_CURRENT_FLEX= "GetCurrentBooking"


MAP_ROUTE_ENDPOINT = {
    AspScript: CommutautoEndpoint.ASP,
    LSIRoute: CommutautoEndpoint.LSI
}

EnumRouteType = Union[AspScript, LSIRoute]


class Cities(Enum):
    MONTREAL = 59
    QUEBEC = 90
    SHERBROOKE = 89
    TROIS_RIVIERES = 110
    GATINEAU = 94
    OTTAWA = 93
    KINGSTON = 97
    SW_ONTARIO = 103
    TORONTO = 105
    HALIFAX = 92
    PARIS = 96
    EDMONTON = 106
    CALGARY = 107

CITIES_MAPPING_STR_ENUM = {
    "montreal": Cities.MONTREAL,
    "quebec": Cities.QUEBEC,
    "sherbrooke": Cities.SHERBROOKE,
    "trois_rivieres": Cities.TROIS_RIVIERES,
    "gatineau": Cities.GATINEAU,
    "ottawa": Cities.OTTAWA,
    "kingston": Cities.KINGSTON,
    "sw_ontario": Cities.SW_ONTARIO,
    "toronto": Cities.TORONTO,
    "halifax": Cities.HALIFAX,
    "paris": Cities.PARIS,
    "edmonton": Cities.EDMONTON,
    "calgary": Cities.CALGARY
}

class Branches(Enum):
    QUEBEC = 1
    ONTARIO = 2
    NOVA_SCOTIA = 3
    FRANCE = 4
    ALBERTA = 10

CitiesToBranchesMapping = {
    Cities.MONTREAL: Branches.QUEBEC,
    Cities.QUEBEC: Branches.QUEBEC,
    Cities.SHERBROOKE: Branches.QUEBEC,
    Cities.TROIS_RIVIERES: Branches.QUEBEC,
    Cities.GATINEAU: Branches.QUEBEC,
    Cities.OTTAWA: Branches.ONTARIO,
    Cities.KINGSTON: Branches.ONTARIO,
    Cities.SW_ONTARIO: Branches.ONTARIO,
    Cities.TORONTO: Branches.ONTARIO,
    Cities.HALIFAX: Branches.NOVA_SCOTIA,
    Cities.PARIS: Branches.FRANCE,
    Cities.EDMONTON: Branches.ALBERTA,
    Cities.CALGARY: Branches.ALBERTA
}

class CarModels(Enum):
    HYUNDAI_VENUE = 1
    CHEVROLET_VOLT = 2
    KIA_NIRO_HYBRID = 3
    KIA_RIO = 4
    TOYOTA_COROLLA_HYBRID = 5
    HYUNDAI_ELANTRA = 6
    TOYOTA_YARIS = 7
    KIA_RIO_EX = 8
    TOYOTA_PRIUS_C = 9
    HYUNDAI_KONA = 10
    TOYOTA_COROLLA = 11
    TOYOTA_PRIUS_V = 12

class CarSizes(Enum):
    COMPACT = 1
    MID_SIZE = 2
    FAMILY_CAR = 3

SIZES_MAPPING_STR_ENUM = {
    "compact": CarSizes.COMPACT,
    "midsize": CarSizes.MID_SIZE,
    "family": CarSizes.FAMILY_CAR
}

CAR_MODELS_MAPPING_STR_ENUM = {
    ('Hyundai', 'Venue'): CarModels.HYUNDAI_VENUE,
    ('Chevrolet', 'Volt'): CarModels.CHEVROLET_VOLT,
    ('Kia', 'Niro hybrid'): CarModels.KIA_NIRO_HYBRID,
    ('Kia', 'Rio'): CarModels.KIA_RIO,
    ('Toyota', 'Corolla hybrid'): CarModels.TOYOTA_COROLLA_HYBRID,
    ('Hyundai', 'Elantra'): CarModels.HYUNDAI_ELANTRA,
    ('Toyota', 'Yaris'): CarModels.TOYOTA_YARIS,
    ('Kia', 'Rio EX'): CarModels.KIA_RIO_EX,
    ('Toyota', 'Prius C'): CarModels.TOYOTA_PRIUS_C,
    ('Hyundai', 'Kona'): CarModels.HYUNDAI_KONA,
    ('Toyota', 'Corolla'): CarModels.TOYOTA_COROLLA,
    ('Toyota', 'Prius V'): CarModels.TOYOTA_PRIUS_V
}

CAR_MODELS_TO_SIZE_MAPPING = {
    CarModels.HYUNDAI_VENUE: CarSizes.COMPACT,
    CarModels.CHEVROLET_VOLT: CarSizes.COMPACT,
    CarModels.KIA_RIO: CarSizes.COMPACT,
    CarModels.TOYOTA_YARIS: CarSizes.COMPACT,
    CarModels.KIA_RIO_EX: CarSizes.COMPACT,
    CarModels.TOYOTA_PRIUS_C: CarSizes.COMPACT,
    CarModels.KIA_NIRO_HYBRID: CarSizes.MID_SIZE,
    CarModels.TOYOTA_COROLLA_HYBRID: CarSizes.MID_SIZE,
    CarModels.HYUNDAI_ELANTRA: CarSizes.MID_SIZE,
    CarModels.HYUNDAI_KONA: CarSizes.MID_SIZE,
    CarModels.TOYOTA_COROLLA: CarSizes.MID_SIZE,
    CarModels.TOYOTA_PRIUS_V: CarSizes.FAMILY_CAR
}


class CarBookingCurrentlyActiveException(BaseException):
    pass

class AlreadyBookedFlexError(BaseException):
    pass

class NoCurrentlyBookedFlex(BaseException):
    pass

class OverlappingWithBookingsWindow(BaseException):
    pass

class ExpiredSearchException(BaseException):
    pass

class AlreadyBookedFlexError(BaseException):
    pass

class NoCurrentlyBookedFlex(BaseException):
    pass

class CarBookingCurrentlyActiveException(BaseException):
    pass

class GetAuthTokenFailure(BaseException):
    pass

class RefreshAuthTokenFailure(BaseException):
    pass

class AuthenticatedClientRequiredException(BaseException):
    pass

class NoStationsWhithinRadiusException(BaseException):
    pass



import logging
import os

# Get the logging level from an environment variable
log_level = os.environ.get('LOG_LEVEL', logging.INFO)

log_file = os.environ.get('LOG_FILE_PATH', "./commutauto.log")
log_format = os.environ.get('LOG_FORMAT', "%(asctime)s - %(name)s - %(levelname)s - %(message)s")

f_handler = logging.FileHandler(log_file)
c_handler = logging.StreamHandler()

logging.basicConfig(
    level=log_level, 
    format=log_format,
    handlers=[f_handler, c_handler])



from typing import List, Dict
import logging
import traceback
from tenacity import RetryError
import math
import logging

from commutauto.client import CommutautoClient
from commutauto.model import Station, Flex, FlexBooking
from commutauto.booking_service.payload import get_payload
from commutauto.booking_service.station import filter_vehicles_whithin_radius

from commutauto.constants import LSIRoute, MAP_ROUTE_ENDPOINT, Cities, Branches
from commutauto.constants import ( FLEX_BASE_WAIT_TIME_SECOND,
                                              FLEX_WAIT_TIME_JITTER_SECOND,
                                              FLEX_EXPIRATION_TIME_SECOND,
                                              MAX_FLEX_EXPIRATION_TIME_SECOND
                                            )
from commutauto.booking_service.retriers import get_available_station_search_retryer
from commutauto.exceptions import AlreadyBookedFlexError, NoCurrentlyBookedFlex, AuthenticatedClientRequiredException

logger = logging.getLogger(__name__)

def get_available_flexes(client: CommutautoClient, branch_id: Branches, city_id: Cities, latitude: float=None, longitude: float=None):
    logger.info(f"Get available flexes. Branch id: {branch_id}, City id: {city_id}, Latitude: {latitude}, Longitude: {longitude}.")
    url_params = get_payload(
        script_name=LSIRoute.GET_FLEX_DISPONIBILITY,
        payload_args={
            "BranchID": branch_id.value,
            "CityID": city_id.value
        }
    )
    response = client.get(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(LSIRoute.GET_FLEX_DISPONIBILITY)],
        script_name=LSIRoute.GET_FLEX_DISPONIBILITY, 
        url_parameters=url_params
    )
    if not response or \
       not "d" in response or \
       not "Vehicles" in response["d"]:
        logger.info("No flexes found.")
        return []
    flexes = parse_flexes(response["d"]["Vehicles"], latitude, longitude)
    logger.debug(f"Flexes found: {flexes}.")
    return flexes

def haversine(lat1, lon1, lat2, lon2):
    R = 6371  # radius of the Earth in kilometers
    lat1, lon1, lat2, lon2 = map(math.radians, [lat1, lon1, lat2, lon2])
    dlat = lat2 - lat1
    dlon = lon2 - lon1
    a = math.sin(dlat/2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon/2)**2
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1-a))
    return R * c

def parse_flexes(flexes_dicts: List[Dict], latitude: float=None, longitude: float=None)-> List[Flex]:
    flexes = []
    for flexes_dict in flexes_dicts:
        flex = None
        if latitude and longitude:
            dist = haversine(
                latitude, longitude, 
                flexes_dict["Latitude"], flexes_dict["Longitude"]
            )
            flex = Flex(**flexes_dict, Distance=dist)
        else:
            flex = Flex(**flexes_dict)
        flexes.append(flex)
    return flexes

def get_available_flexess_whithin_radius(client: CommutautoClient, latitude: float, longitude: float, 
                                radius: float, branch_id: Branches, city_id: Cities, k: int = -1)-> List[Station]:
    """
    Args:
        start_date (dt.datetime): UTC time
        end_date (dt.datetime): UTC time
        radius (float): kilometers
    """
    logger.info(f"Get available flexes whithin radius. Branch id: {branch_id}, City id: {city_id}, Latitude: {latitude}, Longitude: {longitude}, radius: {radius}, k: {k}.")

    if client.authenticated and get_current_flex(client, branch_id=branch_id):
        raise AlreadyBookedFlexError
    
    flexes = get_available_flexes(client, branch_id=branch_id, city_id=city_id, latitude=latitude, longitude=longitude)
    flexes_whithin_radius = filter_vehicles_whithin_radius(flexes, radius, k)
    if flexes_whithin_radius:
        logger.info("Found flexes:")
        for flex in flexes_whithin_radius:
            logger.info(flex)
    return flexes_whithin_radius

def book_flex_from_vin(client: CommutautoClient, car_vin: str, branch_id: Branches) -> int:
    logger.info(f"Book flex from vin. Car vin: {car_vin}, Branch_id: {branch_id}.")

    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    if get_current_flex(client, branch_id=branch_id):
        raise AlreadyBookedFlexError
    
    payload = get_payload(
        script_name=LSIRoute.BOOK_FLEX,
        payload_args={ "CarVIN": car_vin, "BranchID": branch_id.value }
    )

    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(LSIRoute.BOOK_FLEX)],
        script_name=LSIRoute.BOOK_FLEX, 
        payload=payload
    )

    if not response or \
        "d" not in response or \
        "Success" not in response["d"]:
        logger.warn(f"Flex booking failed.")
        return False
    logger.info(f"Successfull flex booking.")
    return response["d"]["Success"]

def cancel_flex_booking(client: CommutautoClient, branch_id: Branches) -> int:
    logger.info(f"Cancel flex booking. Branch_id: {branch_id}.")

    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    if not get_current_flex(client, branch_id=branch_id):
        raise NoCurrentlyBookedFlex
    
    payload = get_payload(
        script_name=LSIRoute.CANCEL_BOOKING_FLEX,
        payload_args={"BranchID": branch_id.value}
    )

    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(LSIRoute.CANCEL_BOOKING_FLEX)],
        script_name=LSIRoute.CANCEL_BOOKING_FLEX, 
        payload=payload
    )

    if not response or \
        not "d" in response or \
        not "Success" in response["d"]:
        logger.warn(f"Flex booking cancellation failed.")
        return False
    logger.info(f"Flex booking cancellation successfull.")
    return response["d"]["Success"]

def get_current_flex(client: CommutautoClient, branch_id: Branches):
    logger.info(f"Check if flex is currently booked.")
    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    payload = get_payload(
        script_name=LSIRoute.GET_CURRENT_FLEX,
        payload_args={"BranchID": branch_id.value}
    )
    response = client.get(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(LSIRoute.GET_CURRENT_FLEX)],
        script_name=LSIRoute.GET_CURRENT_FLEX,
        url_parameters=payload
    )

    if not response or not "d" in response:
        logger.info(f"No flex currently booked.")
        return None
    flex_booking = FlexBooking.from_get_flex_booking_json(response["d"])
    logger.info(f"Flex currently booked {flex_booking}.")
    return flex_booking

def available_flexes_search_with_retries(client: CommutautoClient, 
                                latitude: float, longitude: float,
                                branch_id: Branches, city_id: Cities,
                                radius: float, base_wait_time_second: int=FLEX_BASE_WAIT_TIME_SECOND, 
                                wait_time_jitter_second: int=FLEX_WAIT_TIME_JITTER_SECOND, 
                                expiration_time_second: int=FLEX_EXPIRATION_TIME_SECOND, 
                                longer_pause_after_second: int=None, 
                                longer_pause_max_duration_second: int=None):
    
    logger.info(f"Search flex with retries. Wait time: {base_wait_time_second}s, Jitter: {wait_time_jitter_second}s, Expiration time: {expiration_time_second}.")

    if expiration_time_second > MAX_FLEX_EXPIRATION_TIME_SECOND:
        raise ValueError(f"Retrying for more than {MAX_FLEX_EXPIRATION_TIME_SECOND} is forbidden.")
    
    if client.authenticated and get_current_flex(client, branch_id=branch_id):
        raise AlreadyBookedFlexError
    
    retryer = get_available_station_search_retryer(base_wait_time_second,
                                                   wait_time_jitter_second,
                                                   expiration_time_second,
                                                   longer_pause_after_second,
                                                   longer_pause_max_duration_second)
        
    flex = None
    try:
        flexes = retryer(get_available_flexess_whithin_radius, client, latitude=latitude, longitude=longitude, radius=radius, branch_id=branch_id, city_id=city_id)
        if flexes:
            flex = flexes[0]
    except RetryError:
        logging.exception("Ending retry loop.")
    return flex


from dataclasses import dataclass
import json
from typing import Optional
import logging

logger = logging.getLogger(__name__)

@dataclass
class Station:
    StationID: int
    StationNo: str
    Longitude: float
    Latitude: float
    Distance: float
    StationName: str
    CarID: int
    CarNo: str
    NbrRes: int
    Accessories: int
    lngModel: int
    Brand: str
    Model: str
    Color: str
    CarDesc: str
    HasZone: bool
    EstimatedRangeKm: Optional[float] = None

    def __eq__(self, other):
        return self.Distance == other.Distance
        
    def __lt__(self, other):
        return self.Distance < other.Distance
        
    def __gt__(self, other):
        return self.Distance > other.Distance

    def __str__(self):
        return json.dumps(self.__dict__, indent=4)



import argparse
from datetime import datetime
import logging
from typing import Set

import commutauto.booking_service.station as bs_station
import commutauto.booking_service.flex as bs_flex
from commutauto.client import CommutautoClient
from commutauto.constants import (CITIES_MAPPING_STR_ENUM, Cities, 
                                             MAX_STATION_EXPIRATION_TIME_SECOND,
                                             MAX_FLEX_EXPIRATION_TIME_SECOND,
                                             CitiesToBranchesMapping,
                                             SIZES_MAPPING_STR_ENUM)

logger = logging.getLogger(__name__)

def parse_date(date: str) -> datetime:
    date = datetime.strptime(date, '%Y-%m-%dT%H:%M:%S')
    return date

def search_station(city: str, latitude: float, longitude: float, radius: float, start_date: datetime, 
                   end_date: datetime, search_duration: int=None, book: bool=False, sizes: Set[str]=None, swap: bool = False):
    city = CITIES_MAPPING_STR_ENUM[city]

    if sizes:
        sizes = {SIZES_MAPPING_STR_ENUM[x] for x in sizes}

    client = CommutautoClient(authenticated=False)
    if book:
        client = CommutautoClient(authenticated=True)

    available_station = None
    if not search_duration:
        available_station = bs_station.get_available_stations_whithin_radius(client, 
                                start_date=start_date,
                                end_date=end_date,
                                latitude=latitude,
                                longitude=longitude,
                                city_id=city,
                                radius=radius,
                                allowed_sizes=sizes,
                                swap_bookings=swap,
                                k=1)[0]
    else:
        available_station = bs_station.available_station_search_with_retries(
            client, 
            start_date=start_date, 
            end_date=end_date, 
            latitude=latitude, 
            longitude=longitude, 
            radius=radius,
            city_id=city,
            allowed_sizes=sizes,
            swap_bookings=swap,
            expiration_time_second=search_duration*24*60*60
        )
    
    if not available_station:
        logger.info(f"No station found stopping search.")
        return
    logger.info(f"Station found: {available_station}. Ending search.")

    if not book:
        return
    
    reservation_id = bs_station.book_car_from_id(client, 
                                    start_date=start_date, 
                                    end_date=end_date, 
                                    car_id=available_station.CarID, 
                                    latitude=latitude, 
                                    longitude=longitude,
                                    city_id=city,
                                    swap_bookings=swap)
    if reservation_id:
        logger.info(f"Successfull station booking. Reservation id: {reservation_id}.")
            
def book_station(city: str, latitude: float, longitude: float, start_date: datetime, 
                   end_date: datetime, car_id: str, swap: bool = False):
    city = CITIES_MAPPING_STR_ENUM[city]
    client = CommutautoClient(authenticated=True)

    reservation_id = bs_station.book_car_from_id(client, 
                                    start_date=start_date, 
                                    end_date=end_date, 
                                    car_id=car_id, 
                                    latitude=latitude, 
                                    longitude=longitude,
                                    city_id=city,
                                    swap_bookings=swap)
    if reservation_id:
        logger.info(f"Successfull station booking. Reservation id: {reservation_id}.")

def list_station_bookings():
    client = CommutautoClient(authenticated=True)
    reservations = bs_station.get_car_booking_list(client)

    if not reservations:
        logger.info("No reservation found.")
        return
    
    logger.info(f"Found the following reservations:")
    
    for res in reservations:
        logger.info(res)

def cancel_station_booking(reservation_id: str):
    client = CommutautoClient(authenticated=True)

    reservation = bs_station.get_car_booking(client, reservation_id)

    if not reservation:
        logger.info(f"No reservation found with id {reservation_id}.")
        return
    
    cancel_success = bs_station.cancel_car_booking(client, reservation, latitude=reservation.Latitude, longitude=reservation.Longitude)
    if cancel_success:
        logger.info("Successfull flex booking cancellation.")
    else:
        logger.info("Station booking cancellation failed.")

def search_flex(city: str, latitude: float, longitude: float, radius: float, search_duration: int=None, book: bool=False):
    city = CITIES_MAPPING_STR_ENUM[city]
    branch = CitiesToBranchesMapping[city]

    client = CommutautoClient(authenticated=False)
    if book:
        client = CommutautoClient(authenticated=True)

    available_flex = None
    if not search_duration:
        available_flex = bs_flex.get_available_flexess_whithin_radius(client,
                                                                    branch_id=branch,
                                                                    city_id=city,
                                                                    latitude=latitude, 
                                                                    longitude=longitude, 
                                                                    radius=radius,
                                                                    k=1)[0]
    else:
        available_flex = bs_flex.available_flexes_search_with_retries(client, 
                                latitude=latitude,
                                longitude=longitude,
                                branch_id=branch,
                                city_id=city,
                                radius=radius,
                                expiration_time_second=search_duration*60*60)

    if not available_flex:
        logger.info(f"No flex found stopping search.")
        return
    logger.info(f"Flex found: {available_flex}. Ending search.")

    if not book:
        return
    
    booking_success = bs_flex.book_flex_from_vin(client, 
                                car_vin=available_flex.CarVin,
                                branch_id=branch)
    if booking_success:
        logger.info("Successfull flex booking.")
    else:
        logger.info("Flex booking failed.")

def book_flex(city: str, car_vin: str):
    city = CITIES_MAPPING_STR_ENUM[city]
    branch = CitiesToBranchesMapping[city]
    client = CommutautoClient(authenticated=True)

    booking_success = bs_flex.book_flex_from_vin(client, 
                                car_vin=car_vin,
                                branch_id=branch)
    if booking_success:
        logger.info("Successfull flex booking.")
    else:
        logger.info("Flex booking failed.")

def cancel_flex_booking(city: str):
    city = CITIES_MAPPING_STR_ENUM[city]
    branch = CitiesToBranchesMapping[city]
    client = CommutautoClient(authenticated=True)

    cancel_success = bs_flex.cancel_flex_booking(client, branch_id=branch)
    if cancel_success:
        logger.info("Successfull flex booking cancellation.")
    else:
        logger.info("Flex booking cancellation failed.")
    
# TODO: always assume credentials as env variables? Works whithout .env?
def main():
    parser = argparse.ArgumentParser('Communauto booking manager.')
    subparsers = parser.add_subparsers()

    ###############
    ### STATION ###
    ###############
    station = subparsers.add_parser('station', help='Handle search and booking of station cars.')
    station_parsers = station.add_subparsers()

    ### SEARCH ###
    station_search_parser = station_parsers.add_parser('search', help="Start a station search.")
    station_search_parser.set_defaults(func=search_station)
    station_search_parser.add_argument('--city',
                                       choices=CITIES_MAPPING_STR_ENUM.keys(), 
                                       help='In which city to search for a station car.')
    
    station_search_parser.add_argument('--latitude',
                                       type=float, 
                                       help='Latitude for the center of the search circle. Should be in the selected city.')
    
    station_search_parser.add_argument('--longitude',
                                       type=float, 
                                       help='Longitude for the center of the search circle. Should be in the selected city.')
    
    station_search_parser.add_argument('--radius',
                                       type=float, 
                                       help='Radius of the search circle in kilometers. Only stations whithin the circle will be considered.')
    
    station_search_parser.add_argument('--start_date',
                                       type=parse_date, 
                                       help="""Date at which the car booking will start. Time must be expressed in the timezone of the specified city.
                                               Time will be rounded to closest 15 minutes. Date format must be 'YYYY-MM-DDTHH:mm:ss'.""")

    station_search_parser.add_argument('--end_date',
                                       type=parse_date, 
                                       help="""Date at which the car booking will end. Time must be expressed in the timezone of the specified city.
                                               Time will be rounded to closest 15 minutes. end_date must be at least 30 minutes more than start_date.
                                                Date format must be 'YYYY-MM-DDTHH:mm:ss'.""")
    
    station_search_parser.add_argument('--search_duration', type=float, required=False,
                                       help=f"""Duration of the search in days. If specified, the search will continue until a car is found or 
                                                the search duration is greater than the specified value. If not specified, only one search will be conducted. 
                                                If after this time no car is found, the search will stop. Maximum value is {int(MAX_STATION_EXPIRATION_TIME_SECOND/(24*60*60))} days.""")

    station_search_parser.add_argument('--sizes',
                                       choices=SIZES_MAPPING_STR_ENUM.keys(),
                                       default=None,
                                       nargs='+',
                                       help='What car sizes to allow. Possible values are: compact, midsize and family. If not specified, all sizes are allowed.')
    
    station_search_parser.add_argument('--book', action='store_true', 
                                       help='If the flag is set then if a station is found during the search it will be booked automatically.')
    
    station_search_parser.add_argument('--swap', action='store_true', 
                                       help='If the flag is set and an overlapping booking already exists the search will be enabled. And in the event where a new booking is found, the old one will be deleted.')
    
    ### BOOK ###
    station_book_parser = station_parsers.add_parser('book', help="Book a station.")
    station_book_parser.set_defaults(func=book_station)
    station_book_parser.add_argument('--city',
                                       choices=CITIES_MAPPING_STR_ENUM.keys(), 
                                       help='In which city to search for a station car.')
    
    station_book_parser.add_argument('--latitude',
                                       type=float, 
                                       help='Latitude for the center of the search circle. Should be in the selected city.')
    
    station_book_parser.add_argument('--longitude',
                                       type=float, 
                                       help='Longitude for the center of the search circle. Should be in the selected city.')
    
    station_book_parser.add_argument('--start_date',
                                       type=parse_date, 
                                       help="""Date at which the car booking will start. Time must be expressed in the timezone of the specified city.
                                               Time will be rounded to closest 15 minutes. Date format must be 'YYYY-MM-DDTHH:mm:ss'.""")

    station_book_parser.add_argument('--end_date',
                                       type=parse_date, 
                                       help="""Date at which the car booking will end. Time must be expressed in the timezone of the specified city.
                                               Time will be rounded to closest 15 minutes. end_date must be at least 30 minutes more than start_date.
                                                Date format must be 'YYYY-MM-DDTHH:mm:ss'.""")
    station_book_parser.add_argument('--car_id',
                                       help='Station car id (like 1234).')

    station_book_parser.add_argument('--swap', action='store_true', 
                                       help='If the flag is set and an overlapping booking already exists it will be canceled before creating the new one.')
    
    
    ### CANCEL BOOKING ###
    station_cancel_parser = station_parsers.add_parser('cancel_booking', help="Cancel a station booking.")
    station_cancel_parser.set_defaults(func=cancel_station_booking)
    station_cancel_parser.add_argument('--reservation_id',
                                       help='Station reservation id (like 39762132).')
    
    ### LIST BOOKINGS ###
    station_list_parser = station_parsers.add_parser('list_booking', help="List all bookings.")
    station_list_parser.set_defaults(func=list_station_bookings)
    

    ############
    ### FLEX ###
    ############
    flex = subparsers.add_parser('flex', help='Handle search and booking of flex cars.')
    flex_parsers = flex.add_subparsers()

    ### SEARCH ###
    flex_search_parser = flex_parsers.add_parser('search', help="Start a station search.")
    flex_search_parser.set_defaults(func=search_flex)

    flex_search_parser.add_argument('--city',
                                       choices=CITIES_MAPPING_STR_ENUM.keys(), 
                                       help='In which city to search for a flex car.')
    
    flex_search_parser.add_argument('--latitude',
                                       type=float, 
                                       help='Latitude for the center of the search circle. Should be in the selected city.')
    
    flex_search_parser.add_argument('--longitude',
                                       type=float, 
                                       help='Longitude for the center of the search circle. Should be in the selected city.')
    
    flex_search_parser.add_argument('--radius',
                                       type=float, 
                                       help='Radius of the search circle in kilometers. Only flex cars whithin the circle will be considered.')
    
    flex_search_parser.add_argument('--search_duration', type=int, required=False,
                                       help=f"""Duration of the search in days. If specified, the search will continue until a car is found or 
                                                the search duration is greater than the specified value. If not specified, only one search will be conducted. 
                                                If after this time no car is found, the search will stop. Maximum value is {int(MAX_FLEX_EXPIRATION_TIME_SECOND/(60*60))} days.""")

    flex_search_parser.add_argument('--book', action='store_true', 
                                       help='If the flag is set then if a flex is found during the search it will be booked automatically.')
    
    ### BOOK ###
    flex_book_parser = flex_parsers.add_parser('book', help="Book a station.")
    flex_book_parser.set_defaults(func=book_flex)
    flex_book_parser.add_argument('--city',
                                       choices=CITIES_MAPPING_STR_ENUM.keys(), 
                                       help='In which city to search for a flex car.')
    
    flex_book_parser.add_argument('--car_vin',
                                       help='Vehicle identification number (like JTDLDTB37Z6239023).')

    ### CANCEL BOOKING ###
    flex_cancel_parser = flex_parsers.add_parser('cancel_booking', help="Cancel a station booking.")
    flex_cancel_parser.set_defaults(func=cancel_flex_booking)
    flex_cancel_parser.add_argument('--city',
                                       choices=CITIES_MAPPING_STR_ENUM.keys(), 
                                       help='In which city to search for a flex car.')

    args = vars(parser.parse_args())
    func = args.pop("func")
    func(**args)

if __name__ == "__main__":
    main()


from datetime import datetime
import pytz
import json
import logging

from commutauto.booking_service.date_manager import convert_datetime_to_utc

logger = logging.getLogger(__name__)

class Reservation:
    def __init__(self, ReservationID, CarID, CarNo, CarPlateNumber, StationID, StationName, StationPin, StartDate, StartDateISO, EndDate, EndDateISO, StartDateNextRes, StartDateNextResISO, CarMemoID, CarMemoStationID, FeeType, DestinationID, DestinationName, ReservationStatus, RPTEndDate, CarBranchID, CarAccessories, CarModel, CarDescription, CarEquipment, Accessories, ExpectedNbrKm, Note, LSI, Current, Longitude, Latitude, StationMemo, CarMemo, CreditCardAvailable, CityId, RentalId, Status):
        self.ReservationID = ReservationID
        self.CarID = CarID
        self.CarNo = CarNo
        self.CarPlateNumber = CarPlateNumber
        self.StationID = StationID
        self.StationName = StationName
        self.StationPin = StationPin
        self.StartDateNextRes = StartDateNextRes
        self.StartDateNextResISO = StartDateNextResISO
        self.CarMemoID = CarMemoID
        self.CarMemoStationID = CarMemoStationID
        self.FeeType = FeeType
        self.DestinationID = DestinationID
        self.DestinationName = DestinationName
        self.ReservationStatus = ReservationStatus
        self.RPTEndDate = RPTEndDate
        self.CarBranchID = CarBranchID
        self.CarAccessories = CarAccessories
        self.CarModel = CarModel
        self.CarDescription = CarDescription
        self.CarEquipment = CarEquipment
        self.Accessories = Accessories
        self.ExpectedNbrKm = ExpectedNbrKm
        self.Note = Note
        self.LSI = LSI
        self.Current = bool(Current)
        self.Longitude = Longitude
        self.Latitude = Latitude
        self.StationMemo = StationMemo
        self.CarMemo = CarMemo
        self.CreditCardAvailable = CreditCardAvailable
        self.CityId = CityId
        self.RentalId = RentalId
        self.Status = Status
        
        self.StartDateUTC = datetime.strptime(StartDateISO, '%Y-%m-%dT%H:%M:%S%z').astimezone(pytz.utc)
        self.EndDateUTC = datetime.strptime(EndDateISO, '%Y-%m-%dT%H:%M:%S%z').astimezone(pytz.utc)

        # NOTE: same result can be obtained using local time + gps based timezone inference
        # Don't know which is best. commutauto APIs are not consistant with date format
        # StartDate = datetime.strptime(StartDate, '%m/%d/%Y %I:%M:%S %p')
        # EndDate = datetime.strptime(EndDate, '%m/%d/%Y %I:%M:%S %p')
        # self.StartDateUTC = convert_datetime_to_utc(StartDate, self.Latitude, self.Longitude)
        # self.EndDateUTC = convert_datetime_to_utc(EndDate, self.Latitude, self.Longitude)

    def __str__(self):
        d = {key: value.isoformat() if isinstance(value, datetime) else value for key, value in self.__dict__.items()}
        return json.dumps(d, indent=4)

class ReservationShort:
    def __init__(self, ReservationID, StartDate, EndDate, CarID, CarNo, Station, Longitude, Latitude, Status, Current, Upcoming, LSI, npc, BranchID, RentalID, HasZone):
        self.ReservationID = ReservationID
        self.CarID = CarID
        self.CarNo = CarNo
        self.Station = Station
        self.Longitude = float(Longitude)
        self.Latitude =  float(Latitude)
        self.Status = Status
        self.Current = Current
        self.Upcoming = Upcoming
        self.LSI = LSI
        self.npc = npc
        self.BranchID = BranchID
        self.RentalID = RentalID
        self.HasZone = HasZone

        StartDate = datetime.strptime(StartDate, '%Y/%m/%d %H:%M:%S')
        EndDate = datetime.strptime(EndDate, '%Y/%m/%d %H:%M:%S')
        self.StartDateUTC = convert_datetime_to_utc(StartDate, self.Latitude, self.Longitude)
        self.EndDateUTC = convert_datetime_to_utc(EndDate, self.Latitude, self.Longitude)

    def __str__(self):
        d = {key: value.isoformat() if isinstance(value, datetime) else value for key, value in self.__dict__.items()}
        return json.dumps(d, indent=4)


import datetime as dt
from typing import List, Dict, Union, Set
import heapq
import logging
from tenacity import RetryError
import logging

from commutauto.client import CommutautoClient
from commutauto.model import Station, Reservation, ReservationShort, Flex
from commutauto.booking_service.payload import get_payload
from commutauto.constants import (AspScript, 
                                             MAP_ROUTE_ENDPOINT, 
                                             Cities, 
                                             MINIMUM_SEARCH_RADIUS_KM,
                                             STATION_BASE_WAIT_TIME_SECOND,
                                             STATION_WAIT_TIME_JITTER_SECOND,
                                             STATION_EXPIRATION_TIME_SECOND,
                                             MAX_STATION_EXPIRATION_TIME_SECOND,
                                             CarSizes,
                                             CAR_MODELS_MAPPING_STR_ENUM,
                                             CAR_MODELS_TO_SIZE_MAPPING
                                            )
from commutauto.booking_service.retriers import get_available_station_search_retryer
import commutauto.booking_service.date_manager as dm
from commutauto.exceptions import AuthenticatedClientRequiredException, NoStationsWhithinRadiusException, OverlappingWithBookingsWindow

logger = logging.getLogger(__name__)

def parse_stations(station_dicts: List[Dict])-> List[Station]:
    stations = []
    for station_dict in station_dicts:
        stations.append(Station(**station_dict))
    return stations

def filter_available_stations(stations: List[Station]) -> List[Station]:
    available_stations = []
    for station in stations:
        if station.NbrRes == 0 or station.NbrRes == "0":
            available_stations.append(station)
    return available_stations

def filter_closest_station(stations: List[Station]) -> Station:
    min_distance = float("inf")
    closest_station = None
    for s in stations:
        if s.Distance < min_distance:
            min_distance = s.Distance
            closest_station = s
    return closest_station

def filter_car_size_station(stations: List[Station], allowed_sizes: Set[CarSizes]) -> Station:
    logger.info(f"Filter vehicles based on vehice size. Allowed sizes are: {allowed_sizes}.")
    allowed_stations = []
    for station in stations:
        model = CAR_MODELS_MAPPING_STR_ENUM[(station.Brand, station.Model)]
        car_size = CAR_MODELS_TO_SIZE_MAPPING[model]
        if car_size in allowed_sizes:
            allowed_stations.append(station)
    return allowed_stations

def check_stations_whithin_radius(stations: List[Station], radius: float):
    closest_station = filter_closest_station(stations)
    if closest_station and closest_station.Distance > radius:
        err_msg = f"Closest station is located at {closest_station.Distance} kilometers from the provided "
        err_msg += f"GPS coordinates. The specified search radius '{radius} kilometers' is lower than this distance."
        raise NoStationsWhithinRadiusException(err_msg)

def filter_vehicles_whithin_radius(vehicles: List[Union[Station, Flex]], radius: int, k: int = -1) -> List[Union[Station, Flex]]:
    logger.info(f"Filter vehicles based on radius {radius}km with k {k}.")

    if radius < MINIMUM_SEARCH_RADIUS_KM:
        raise ValueError(f"Search radius '{radius}' is lower than minimum allowed radius value '{MINIMUM_SEARCH_RADIUS_KM}'.")
    
    vehicles_whithin_radius = []
    minheap = []
    for vehicle in vehicles:
        heapq.heappush(minheap, vehicle)

    while minheap:
        if k != -1 and len(vehicles_whithin_radius) >= k:
            break
        vehicle = heapq.heappop(minheap)
        if vehicle.Distance > radius:
            break
        vehicles_whithin_radius.append(vehicle)
    return vehicles_whithin_radius

def prepare_and_validate_time_window(start_date: dt.datetime, 
                                end_date: dt.datetime, latitude: float, longitude: float):
    start_date = dm.convert_datetime_to_utc(start_date, latitude, longitude)
    end_date = dm.convert_datetime_to_utc(end_date, latitude, longitude)
    start_date, end_date = dm.round_booking_time_window(start_date, end_date)
    dm.validate_booking_request_time_window(start_date, end_date)
    dm.check_start_date_greater_than_quarter_hour_from_now(start_date)
    dm.check_start_date_sooner_than_one_month(start_date)
    return start_date, end_date

def check_existing_booking_overlap(client: CommutautoClient, start_date: dt.datetime, end_date: dt.datetime) -> List[ReservationShort]:
    reservations = get_car_booking_list(client)
    overlapping_reservations = dm.check_no_booking_intersection_with_search_window(start_date, end_date, reservations)
    return overlapping_reservations

def raise_reservation_overlap_error(overlapping_reservations: List[ReservationShort], start_date: dt.datetime, end_date: dt.datetime):
    err_msg = f"Provided search time window '[{start_date};{end_date}]' overlaps with the following reservations: \n"
    for res in overlapping_reservations:
        err_msg += f"{res} \n"
    raise OverlappingWithBookingsWindow(err_msg)

def get_stations(client: CommutautoClient, start_date: dt.datetime, 
                                end_date: dt.datetime, latitude: float, longitude: float, city_id: Cities, swap_bookings: bool = False):
    logger.info(f"Get available stations. City id: {city_id}, Latitude: {latitude}, Longitude: {longitude}, Start: {start_date}, End: {end_date}.")
    start_date, end_date = prepare_and_validate_time_window(start_date, end_date, latitude, longitude)

    if client.authenticated:
        overlaping_res = check_existing_booking_overlap(client, start_date, end_date)
        if overlaping_res and not swap_bookings:
            raise_reservation_overlap_error(overlaping_res, start_date, end_date)

    payload = get_payload(
        script_name=AspScript.GET_CAR_DISPONIBILITY,
        payload_args={
            "StartDateUTC": start_date, 
            "EndDateUTC": end_date, 
            "Latitude": latitude,
            "Longitude": longitude,
            "CityID": city_id.value
        }
    )
    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(AspScript.GET_CAR_DISPONIBILITY)],
        script_name=AspScript.GET_CAR_DISPONIBILITY, 
        payload=payload
    )
    if not response or "Stations" not in response:
        logger.info("No stations found.")
        return []
    
    stations = parse_stations(response["Stations"])
    logger.debug(f"Stations found: {stations}.")
    return stations

def get_available_stations_whithin_radius(client: CommutautoClient, start_date: dt.datetime, 
                                end_date: dt.datetime, latitude: float, longitude: float, city_id: Cities,
                                radius: float, k: int = -1, allowed_sizes: Set[CarSizes] = None, swap_bookings: bool = False)-> List[Station]:
    """
    Args:
        start_date (dt.datetime): UTC time
        end_date (dt.datetime): UTC time
        radius (float): kilometers
    """
    logger.info(f"Get available stations whithin radius {radius}.")

    stations = get_stations(client, start_date, end_date, latitude, longitude, city_id=city_id, swap_bookings=swap_bookings)
    check_stations_whithin_radius(stations, radius)
    available_stations = filter_available_stations(stations)
    stations_whithin_radius = filter_vehicles_whithin_radius(available_stations, radius, k)

    if allowed_sizes:
        stations_whithin_radius = filter_car_size_station(stations_whithin_radius, allowed_sizes)

    if stations_whithin_radius:
        logger.info("Found stations:")
        for station in stations_whithin_radius:
            logger.info(station)
    return stations_whithin_radius

def book_car_from_id(client: CommutautoClient, start_date: dt.datetime, 
                                end_date: dt.datetime, car_id: int, latitude: float, longitude: float, city_id: Cities, swap_bookings: bool = False) -> int:
    logger.info(f"Book station from id. Car id: {car_id}, City id: {city_id}, Latitude: {latitude}, Longitude: {longitude}, Start: {start_date}, End: {end_date}.")
    
    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    start_date, end_date = prepare_and_validate_time_window(start_date, end_date, latitude, longitude)

    if client.authenticated:
        overlaping_res = check_existing_booking_overlap(client, start_date, end_date)
        if overlaping_res:
            if not swap_bookings:
                raise_reservation_overlap_error(overlaping_res, start_date, end_date)
            else:
                for res in overlaping_res:
                    complete_res = get_car_booking(client, res.ReservationID)
                    cancel_car_booking(client, complete_res, latitude, longitude)
        
    payload = get_payload(
        script_name=AspScript.BOOK_CAR,
        payload_args={
            "StartDateUTC": start_date, 
            "EndDateUTC": end_date, 
            "CarId": car_id,
            "Latitude": latitude,
            "Longitude": longitude,
            "CityID": city_id.value
        }
    )

    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(AspScript.BOOK_CAR)],
        script_name=AspScript.BOOK_CAR, 
        payload=payload
    )
    if not response or \
        not "data" in response or \
        not "ReservationID" in response["data"] or \
        not(response["data"]["ReservationID"] > 0 and \
            response["data"]["Error"] == 0):
        logger.warn(f"Station booking failed.")
        return None
    logger.info(f"Successfull station booking.")
    return response["data"]["ReservationID"]

def get_car_booking(client: CommutautoClient, reservation_id: int) -> Reservation:
    logger.info(f"Get station booking. Reservation id: {reservation_id}.")

    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    payload = get_payload(
        script_name=AspScript.GET_BOOKING,
        payload_args={
            "ReservationID": reservation_id
        }
    )
    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(AspScript.GET_BOOKING)],
        script_name=AspScript.GET_BOOKING, 
        payload=payload
    )

    if not response or "Reservation" not in response:
        logger.info("No reservation found.")
        return None

    reservation = Reservation(**response["Reservation"])
    logger.info(f"Found reservation {reservation}.")
    return reservation

def cancel_car_booking(client: CommutautoClient, reservation: Reservation, latitude: float, longitude: float) -> bool:
    logger.info(f"Cancel station booking. Reservation: {reservation}, Latitude {latitude}, longitude {longitude}.")
    
    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    payload = get_payload(
        script_name=AspScript.CANCEL_BOOKING,
        payload_args={
            "StartDateUTC": reservation.StartDateUTC,
            "EndDateUTC": reservation.EndDateUTC,
            "ReservationID": reservation.ReservationID,
            "NewFeeType": reservation.FeeType,
            "ExpectedNbrKm": reservation.ExpectedNbrKm,
            "Latitude": latitude,
            "Longitude": longitude
        }
    )
    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(AspScript.CANCEL_BOOKING)],
        script_name=AspScript.CANCEL_BOOKING, 
        payload=payload)

    if not response or \
        not "data" in response or \
        not "ReservationID" in response["data"] or \
        not(response["data"]["ReservationID"] > 0 and response["data"]["Error"] == 0):
        logger.warn(f"Station booking cancellation failed.")
        return False
    logger.info(f"Station booking cancellation successfull.")
    return True

def parse_reservations(reservation_dicts: List[Dict])-> List[ReservationShort]:
    reservations = []
    for reservation_dict in reservation_dicts:
        reservations.append(ReservationShort(**reservation_dict))
    return reservations

def get_car_booking_list(client: CommutautoClient)-> List[ReservationShort]:
    logger.info(f"Get station booking list.")

    if not client.authenticated:
        raise AuthenticatedClientRequiredException()
    
    payload = get_payload(
        script_name=AspScript.GET_BOOKING_LIST,
        payload_args={}
    )
    response = client.post(
        commutauto_endpoint=MAP_ROUTE_ENDPOINT[type(AspScript.GET_BOOKING_LIST)],
        script_name=AspScript.GET_BOOKING_LIST, 
        payload=payload)
    
    if not response or "data" not in response:
        logger.info(f"No station bookings..")   
        return []
    reservations = parse_reservations(response["data"])
    if reservations:
        logger.info("Found station bookings:")
        for res in reservations:
            logger.info(res)
    return reservations

def available_station_search_with_retries(client: CommutautoClient, start_date: dt.datetime, 
                                end_date: dt.datetime, latitude: float, longitude: float, city_id: Cities, 
                                radius: float, allowed_sizes: Set[CarSizes] = None, swap_bookings: bool = False,
                                base_wait_time_second: int=STATION_BASE_WAIT_TIME_SECOND, 
                                wait_time_jitter_second: int=STATION_WAIT_TIME_JITTER_SECOND, 
                                expiration_time_second: int=STATION_EXPIRATION_TIME_SECOND, 
                                longer_pause_after_second: int=None, 
                                longer_pause_max_duration_second: int=None):
    logger.info(f"Search stations with retries. Wait time: {base_wait_time_second}s, Jitter: {wait_time_jitter_second}s, Expiration time: {expiration_time_second}.")
    
    if expiration_time_second > MAX_STATION_EXPIRATION_TIME_SECOND:
        raise ValueError(f"Retrying for more than {MAX_STATION_EXPIRATION_TIME_SECOND} is forbidden.")
    
    retryer = get_available_station_search_retryer(base_wait_time_second,
                                                   wait_time_jitter_second,
                                                   expiration_time_second,
                                                   longer_pause_after_second,
                                                   longer_pause_max_duration_second)
        
    station = None
    try:
        stations = retryer(get_available_stations_whithin_radius, client, start_date=start_date, 
                           end_date=end_date, latitude=latitude, longitude=longitude, radius=radius, city_id=city_id, 
                           allowed_sizes=allowed_sizes, swap_bookings=swap_bookings)
        if stations:
            station = stations[0]
    except RetryError:
        logging.exception("Ending retry loop.")
    return station


import io
import os
from pathlib import Path

from setuptools import find_packages, setup

# Package meta-data.
NAME = 'commutauto'
DESCRIPTION = 'CLI to interact with your car sharing service API. Search for available vehicles, manage your bookings and more...'
URL = 'https://github.com/elbuco1/commutauto'
EMAIL = 'lrtboucaud@gmail.com'
AUTHOR = "Laurent Boucaud"
REQUIRES_PYTHON = '>=3.7'

def list_reqs(fname='requirements.txt'):
    with open(fname) as fd:
        return fd.read().splitlines()

INSTALL_REQUIRES = list_reqs()

TEST_REQUIRES = [
    'pytest',
    'tox'
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# try:
#     with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
#         long_description = '\n' + f.read()
# except FileNotFoundError:
#     long_description = DESCRIPTION

# import version
ROOT_DIR = Path(__file__).resolve().parent
PACKAGE_DIR = ROOT_DIR / NAME

about = {}
with open(PACKAGE_DIR / 'VERSION') as f:
    _version = f.read().strip()
    about['__version__'] = _version

setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    package_data={
        'commutauto': ['VERSION']
    },
    install_requires=INSTALL_REQUIRES,
    extras_require={
        "tests": TEST_REQUIRES
    },
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
    ],
    entry_points={
        'console_scripts': [
            'commutauto=commutauto.__main__:main'
        ]
    }
)



from typing import Optional
import datetime
import json
import re
import logging

logger = logging.getLogger(__name__)

class FlexBooking:
    def __init__(self, BookingId: int, BookingStatus: int, CarId: int, CarModel: str,
                 CarNo: int, CarPlate: str, CarVin: str, CityID: int, CreditCardAvailable: bool,
                 EndDateUTC: datetime.datetime, EnergyLevel: Optional[float], InUse: bool, IsElectric: bool,
                 LastUse: int, LastUseDateUTC: datetime.datetime, Latitude: float, Longitude: float,
                 RentalId: int, ReservationStartDateUTC: Optional[datetime.datetime], StartDateUTC: datetime.datetime,
                 VehiclePromotions: Optional[str], CurrentDateUTC: datetime.datetime) -> None:
        self.BookingId = BookingId
        self.BookingStatus = BookingStatus
        self.CarId = CarId
        self.CarModel = CarModel
        self.CarNo = CarNo
        self.CarPlate = CarPlate
        self.CarVin = CarVin
        self.CityID = CityID
        self.CreditCardAvailable = CreditCardAvailable
        self.EndDateUTC = EndDateUTC
        self.EnergyLevel = EnergyLevel
        self.InUse = InUse
        self.IsElectric = IsElectric
        self.LastUse = LastUse
        self.LastUseDateUTC = LastUseDateUTC
        self.Latitude = Latitude
        self.Longitude = Longitude
        self.RentalId = RentalId
        self.ReservationStartDateUTC = ReservationStartDateUTC
        self.StartDateUTC = StartDateUTC
        self.VehiclePromotions = VehiclePromotions
        self.CurrentDateUTC = CurrentDateUTC

    @classmethod
    def from_get_flex_booking_json(cls, json_data):
        if "Booking" not in json_data or "UtcDates" not in json_data:
            return None
        
        booking_dict, utc_dates = json_data["Booking"], json_data["UtcDates"]
        del booking_dict["EndDate"]
        del booking_dict["StartDate"]
        del booking_dict["LastUseDate"]
        del booking_dict["ReservationStartDate"]
        del booking_dict["__type"]

        booking_dict["CurrentDateUTC"] = parse_iso_with_milliseconds_to_utc_datetime(utc_dates["CurrentDate"])
        booking_dict["EndDateUTC"] = parse_iso_with_milliseconds_to_utc_datetime(utc_dates["EndDate"])
        booking_dict["StartDateUTC"] = parse_iso_with_milliseconds_to_utc_datetime(utc_dates["StartDate"])
        booking_dict["LastUseDateUTC"] = parse_iso_with_milliseconds_to_utc_datetime(utc_dates["LastUseDate"])
        booking_dict["ReservationStartDateUTC"] = parse_iso_with_milliseconds_to_utc_datetime(utc_dates["ReservationStartDate"])

        flex_booking = cls(**booking_dict)
        return flex_booking
    def __str__(self):
        d = {key: value.isoformat() if isinstance(value, datetime.datetime) else value for key, value in self.__dict__.items()}
        return json.dumps(d, indent=4)
    
def parse_iso_with_milliseconds_to_utc_datetime(date: str)-> datetime.datetime:
    # remove  millisecond "\"2023-03-03T01:57:02.495417+00:00\"" (before the +)
    if date == "null":
        return None
    date = date.strip("\"")
    date = re.sub(r'\.\d+', '', date)
    isodate= datetime.datetime.fromisoformat(date)
    return isodate


import json
import secrets
import hashlib
import base64
import requests
from requests.cookies import RequestsCookieJar
from bs4 import BeautifulSoup
from typing import Tuple, Callable
import os
from dotenv import load_dotenv
import logging

from commutauto.client.auth_token import AuthToken
from commutauto.booking_service.payload import Payload
from commutauto.constants import CommutautoEndpoint, EnumRouteType
from commutauto.exceptions import GetAuthTokenFailure, RefreshAuthTokenFailure

logger = logging.getLogger(__name__)

class CommutautoClient():
    def __init__(self, 
            commutauto_security_api_endpoint = "https://securityservice.reservauto.net",
            authentication_header_value = "Basic Q29tbXVuYXV0b01vYmlsZUFwcDI6dUVWbEM0dzdiUkMyWlZtQWRjWG9xeXUxdQ==", 
            token_file_path=os.environ.get('COMMUTAUTO_TOKEN_FILE', "token.json"),
            authenticated=True):
        
        logger.info(f"Create CommutautoClient. Authentication is {'enabled' if authenticated else 'disabled'}.")

        load_dotenv()
        self.token_file_path = token_file_path
        self.commutauto_security_api_endpoint = commutauto_security_api_endpoint
        self.authentication_header_value = authentication_header_value
        self.authenticated = authenticated
        if self.authenticated:
            self.load_token_from_json()
    
    def load_token_from_json(self):
        try:
            logger.info(f"Load token from {self.token_file_path}.")
            with open(self.token_file_path) as f:
                self.token = AuthToken(**json.load(f))
        except FileNotFoundError:
            self.token = None
            logger.exception("Token file not found. Defaulting to None.")

    def get_token(self):
        logger.info(f"Get authentication token from API.")

        code_verifier = generate_random_code_verifier()
        code_challenge = encode_code_verifier_to_code_challenge(code_verifier)
        
        antiforgery_token, antiforgery_cookies, \
            return_url = get_and_parse_authentication_form(self.commutauto_security_api_endpoint, code_challenge)

        redirect_url, authentication_cookies = \
            post_authentication_form_and_parse_response(self.commutauto_security_api_endpoint, antiforgery_token,  antiforgery_cookies, return_url)
        
        redirect_oidc_uri, oidc_code = get_oidc_information(redirect_url, authentication_cookies)

        token = get_authentication_token(self.commutauto_security_api_endpoint, self.authentication_header_value, redirect_oidc_uri, oidc_code, code_verifier)
        
        if not token:
            raise GetAuthTokenFailure

        token.save_json(self.token_file_path)
        self.load_token_from_json()

    def refresh_token(self):
        logging.info(f"Refresh token.")

        if self.token:
            logging.info(f"Token found. Refreshing...")
            refresh_token_token = self.token.refresh_token
            token = refresh_token(self.commutauto_security_api_endpoint, self.authentication_header_value, refresh_token_token)

            if not token:
                raise RefreshAuthTokenFailure

            token.save_json(self.token_file_path)
            self.load_token_from_json()

    def authenticate(self):
        logging.info(f"Authenticate client.")

        if not self.token:
            self.get_token()
        elif self.token.is_expired():
            self.refresh_token()

    def get_token_header(self) -> dict:
        if not self.authenticated:
            return {}
        
        self.authenticate()
        token_header = {
            "authorization": f"{self.token.token_type} {self.token.access_token}"
        }
        return token_header

    def get(self, commutauto_endpoint: CommutautoEndpoint, script_name: EnumRouteType, url_parameters = Payload, headers: dict = None) -> dict:
        url = os.path.join(commutauto_endpoint.value, script_name.value)
        if headers is None:
            headers = {}
        headers.update(self.get_token_header())
        args = {
            "url": url,
            "headers": headers,
            "params": url_parameters.to_dict()
        }
        response = handle_requests_exceptions(requests.get, args)
        log_response(response)
        return parse_response_content(response) if response else None

    def post(self, commutauto_endpoint: CommutautoEndpoint, script_name: EnumRouteType, payload: Payload, headers: dict = None) -> dict:
        url = os.path.join(commutauto_endpoint.value, script_name.value)
        if headers is None:
            headers = {}
        headers.update(self.get_token_header())
        args = {
            "url": url,
            "headers": headers
        }
        
        payload_arg_name = "json"
        if payload.is_url_encoded():
            payload_arg_name = "data"
        args[payload_arg_name] = payload.to_dict()

        response = handle_requests_exceptions(requests.post, args)
        log_response(response)
        return parse_response_content(response) if response else None

def parse_response_content(response: requests.Response) -> dict:
    return json.loads(response.content.decode('utf-8'))

def handle_requests_exceptions(requests_method: Callable, args: dict) -> requests.Response:
    try:
        response = requests_method(**args)
        response.raise_for_status()
    except Exception:
        logger.exception("Exception occurred when sending request with Commutauto client.")
        return None
    return response

def get_authentication_token(commutauto_security_api_endpoint: str, authentication_header_value: str, redirect_oidc_uri: str, oidc_code: str, code_verifier: str) -> AuthToken:
    body = {
        "code": oidc_code,
        "grant_type": "authorization_code",
        "redirect_uri": redirect_oidc_uri,
        "code_verifier": code_verifier
    }
    headers = {"authorization": authentication_header_value }

    logging.info(f"Send get token http request.")
    logging.info(f"Headers: {headers}.")
    logging.info(f"Body: {body}.")

    response = requests.post(
        get_ask_token_url(commutauto_security_api_endpoint),
        data=body,
        headers=headers,
        stream=True,
        allow_redirects=False
        )
    log_response(response)

    if response.status_code != 200:
        return None

    token = extract_token_from_response(response)
    return token

def refresh_token(commutauto_security_api_endpoint: str, authentication_header_value: str, refresh_token_token: str) -> AuthToken:
    body = {
        "refresh_token": refresh_token_token,
        "grant_type": "refresh_token"
    }
    headers = {"authorization": authentication_header_value }
    response = requests.post(
        get_ask_token_url(commutauto_security_api_endpoint),
        data=body,
        headers=headers,
        stream=True,
        allow_redirects=False
        )
    if response.status_code != 200:
        return None

    token = extract_token_from_response(response)
    return token

def extract_token_from_response(response: requests.Response) -> AuthToken:
    token = AuthToken(**json.loads(response.content.decode('utf-8')))
    return token

def get_oidc_information(redirect_url: str, authentication_cookies: str) -> Tuple[str, str]:
    logger.info(f"Get oidc information.")

    response = requests.get(
            redirect_url,
            cookies=authentication_cookies,
            stream=True,
            allow_redirects=False
        )
    log_response(response)
    redirect_oidc_uri, oidc_code = parse_oidc_header_location(response.headers['location'])
    logger.info(f"Redirect oidc uri: {redirect_oidc_uri}.")
    logger.info(f"Redirect oidc code: {oidc_code}.")

    return redirect_oidc_uri, oidc_code

def parse_oidc_header_location(redirect_oidc_location: str) -> Tuple[str, str]:
    redirect_oidc_uri = redirect_oidc_location.split("#")[0]
    oidc_code = redirect_oidc_location.split("#")[1] \
                                .split("&")[0] \
                                .split("code=")[1]
    return redirect_oidc_uri, oidc_code

def post_authentication_form_and_parse_response(
        commutauto_security_api_endpoint: str, antiforgery_token: str, 
        antiforgery_cookies: RequestsCookieJar, return_url: str) -> Tuple[str, RequestsCookieJar]:
    logger.info(f"Post authentication form.")
    
    post_form_body = {
            "ReturnUrl": return_url,
            "ClientId": "CommunautoMobileApp2",
            "BranchId": "Communauto_Quebec",
            "Username": os.environ.get('USER_NAME'),
            "Password": os.environ.get('PASSWORD'),
            "__RequestVerificationToken": antiforgery_token,
            "RecaptchaResponse": '""'
        }

    response = requests.post(
        get_post_form_url(commutauto_security_api_endpoint),
        data=post_form_body,
        cookies=antiforgery_cookies,
        stream=True,
        allow_redirects=False
    )
    log_response(response)

    redirect_url = f"{commutauto_security_api_endpoint}{response.headers['location']}"
    authentication_cookies = response.cookies
    return redirect_url, authentication_cookies

def log_response(response: requests.Response):
    logger.debug(f"Response status code: {response.status_code}.")
    logger.debug(f"Response headers: {response.headers}.")
    logger.debug(f"Response content: {response.content}.")
    logger.debug(f"Response cookies: {response.cookies}.")

def get_and_parse_authentication_form(commutauto_security_api_endpoint: str, code_challenge: str) -> Tuple[str, RequestsCookieJar, str]:
    logger.info(f"Get authentication form.")
    ask_form_url = get_ask_form_url(commutauto_security_api_endpoint, code_challenge)
    response = requests.get(ask_form_url, allow_redirects=True)
    soup = BeautifulSoup(response.text, 'html.parser')
    antiforgery_token = get_antiforgery_token_from_form(soup)
    return_url = get_return_url_from_form(soup)
    antiforgery_cookies = response.cookies
    logger.info(f"Antiforgery token: {antiforgery_token}.")
    logger.info(f"Antiforgery cookies: {antiforgery_cookies}.")
    logger.info(f"Return Url: {return_url}")
    return antiforgery_token, antiforgery_cookies, return_url

def get_antiforgery_token_from_form(soup: BeautifulSoup) -> str:
    form = soup.find('form')
    request_verification_token = form.find('input', {'name': '__RequestVerificationToken'})["value"]
    return request_verification_token

def get_return_url_from_form(soup: BeautifulSoup) -> str:
    form = soup.find('form')
    return_url = form.find('input', {"name": "ReturnUrl"})["value"]
    return return_url

def get_ask_form_url(commutauto_security_api_endpoint: str, code_challenge: str) -> str:
    return f"{commutauto_security_api_endpoint}/connect/authorize?client_id=CommunautoMobileApp2&redirect_uri=com.communauto.reservauto%3A%2Foidcredirect&scope=openid%20offline_access%20communautorestapi%20reservautofrontofficerestapi&response_type=code%20id_token&state=851ed6f4-c660-1fd6-19e9-143465bdec5c-72194b-ffdd480cabac-47dd&nonce=49ec9137-761a-9832-1536-9dcb27df0c30-6c4238-3fd9f84fce5f-d8cf&code_challenge={code_challenge}&code_challenge_method=S256&branch_id=1&prompt=login&hideTopHeader=true&ui_locales=en-CA"

def get_post_form_url(commutauto_security_api_endpoint: str) -> str:
    return f"{commutauto_security_api_endpoint}/Account/Login"

def get_ask_token_url(commutauto_security_api_endpoint: str) -> str:
    return f"{commutauto_security_api_endpoint}/connect/token"

def generate_random_code_verifier(segment_lengths=[8, 4, 4, 4, 12, 6, 12, 4]):
    # Generate a random string of the desired length (in bytes)
    num_bytes = int(sum(segment_lengths) / 2) + 1
    rand_bytes = secrets.token_bytes(num_bytes)

    # Convert the bytes to a hex string
    hex_str = rand_bytes.hex()

    # Insert hyphens between segments
    segments = []
    start_index = 0
    for length in segment_lengths:
        segments.append(hex_str[start_index:start_index+length])
        start_index += length
    random_code = '-'.join(segments)
    logger.info(f"Generated random code verifier: {random_code}.")
    return random_code

def encode_code_verifier_to_code_challenge(code_verifier):
    # Calculate the SHA256 hash of the code verifier
    code_challenge = hashlib.sha256(code_verifier.encode()).digest()
    # Base64 URL encode the hash
    code_challenge = base64.urlsafe_b64encode(code_challenge).rstrip(b'=')
    # Convert the bytes to a string
    code_challenge = code_challenge.decode()
    logger.info(f"Encoded code verifier {code_verifier} to code challenge {code_challenge}.")
    return code_challenge

if __name__ == "__main__":
    load_dotenv()
    client = CommutautoClient()
    print(client.get_token_header())


from tenacity import (Retrying,
                      RetryCallState,
                      retry_if_result,
                      stop_after_delay
                      )
from tenacity.wait import wait_base
import time
import random
import logging

logger = logging.getLogger(__name__)

class wait_random_plus_long_pauses(wait_base):
    """Wait strategy that retries every base_wait_time_second + random.uniform(0, wait_time_jitter_second).
    And every longer_pause_after does a longer pause of random.uniform(0, longer_pause_max_duration_second)*60
    """
    def __init__(self, base_wait_time_second: int, 
                       wait_time_jitter_second: float, 
                       longer_pause_after_second: int = None,
                       longer_pause_max_duration_second: int = None) -> None:
        if (not longer_pause_after_second) ^ (not longer_pause_max_duration_second):
            raise ValueError("longer_pause_after_second is set but longer_pause_max_duration_second is not set")
        
        self.base_wait_time_second = base_wait_time_second
        self.wait_time_jitter_second = wait_time_jitter_second
        self.longer_pause_after_second = longer_pause_after_second
        self.longer_pause_max_duration_second = longer_pause_max_duration_second
        self.n_longer_pause = 1

    def __call__(self, retry_state: RetryCallState) -> float:
        wait_time = self.base_wait_time_second + random.uniform(0, self.wait_time_jitter_second)
        elapsed_time = time.monotonic() - retry_state.start_time
        if self.longer_pause_after_second and \
            (elapsed_time - (self.n_longer_pause*self.longer_pause_after_second)) > 0:
            wait_time = random.uniform(0, self.longer_pause_max_duration_second)
            logger.info(f'Taking a longer break. Elapsed time: {elapsed_time}. This is the {self.n_longer_pause}th longer break so far')
            self.n_longer_pause += 1
        logger.info(f'Will wait for {wait_time} seconds.')
        return wait_time
    
def get_available_station_search_retryer(base_wait_time_second: int, 
                                   wait_time_jitter_second: int, 
                                   expiration_time_second: int, 
                                   longer_pause_after_second: int, 
                                   longer_pause_max_duration_second: int) -> Retrying:
    retryer = Retrying(
        retry=retry_if_result(lambda x: len(x) == 0),
        wait=wait_random_plus_long_pauses(base_wait_time_second, wait_time_jitter_second, longer_pause_after_second, longer_pause_max_duration_second),
        stop=stop_after_delay(expiration_time_second),
        after=log_total_elapsed_time
    )
    return retryer

def log_total_elapsed_time(retry_state):
    elapsed_time = time.monotonic() - retry_state.start_time
    logger.info(f'Total elapsed time: {elapsed_time:.2f} seconds')

def available_station_search_test(success_proba=0.05):
    stations_whithin_radius = []
    if random.uniform(0, 1) > 1 - success_proba:
        stations_whithin_radius.append(1)
    return stations_whithin_radius


